import base64
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from rest_framework.decorators import api_view, permission_classes
from location.models import LocationManager
from report.services import ReportService
from core.security import checkUserWithRights
from .services import ClaimReportService
from .reports import claim
from .apps import ClaimConfig
from .models import ClaimAttachment
from django.utils.translation import gettext as _
import core


@api_view(['GET'])
def print(request):
    if not request.user.has_perms(ClaimConfig.claim_print_perms):
        raise PermissionDenied(_("unauthorized"))
    report_service = ReportService(request.user)
    report_data_service = ClaimReportService(request.user)
    data = report_data_service.fetch(request.GET['uuid'])
    return report_service.process('claim_claim', data, claim.template)


@api_view(["GET", "POST"])
@permission_classes(
    [
        checkUserWithRights(
            ClaimConfig.gql_query_claims_perms,
        )
    ]
)
def attach(request):
    queryset = ClaimAttachment.objects.filter(*core.filter_validity())
    if settings.ROW_SECURITY:
        from location.models import LocationManager
        queryset = LocationManager().build_user_location_filter_query(request.user._u, prefix='health_facility__location',
                                                                      queryset=queryset.select_related("claim"), loc_types=['D'])
    attachment = queryset\
        .filter(id=request.GET['id'])\
        .first()
    if not attachment:
        raise PermissionDenied(_("unauthorized"))

    if ClaimConfig.claim_attachments_root_path and attachment.url is None:
        response = HttpResponse(status=404)
        return response

    if not ClaimConfig.claim_attachments_root_path and attachment.document is None:
        response = HttpResponse(status=404)
        return response

    response = HttpResponse(content_type=("application/x-binary" if attachment.mime is None else attachment.mime))
    response['Content-Disposition'] = 'attachment; filename=%s' % attachment.filename
    if ClaimConfig.claim_attachments_root_path:
        f = open('%s/%s' % (ClaimConfig.claim_attachments_root_path, attachment.url), "rb")
        response.write(f.read())
        f.close()
    else:
        response.write(base64.b64decode(attachment.document))
    return response



# Generated by Django 4.2.10 on 2024-03-01 08:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0028_claimserviceitem_claimserviceservice'),
    ]

    operations = [
        migrations.RenameField(
            model_name='claimserviceitem',
            old_name='pcpDate',
            new_name='created_date',
        ),
        migrations.RenameField(
            model_name='claimserviceitem',
            old_name='idCsi',
            new_name='id',
        ),
        migrations.RenameField(
            model_name='claimserviceservice',
            old_name='pcpDate',
            new_name='created_date',
        ),
        migrations.RenameField(
            model_name='claimserviceservice',
            old_name='idCss',
            new_name='id',
        ),
        
        migrations.RenameField(
            model_name='claimserviceservice',
            old_name='claimlinkedService',
            new_name='claim_service',
        ),
        migrations.RenameField(
            model_name='claimserviceitem',
            old_name='claimlinkedItem',
            new_name='claim_service',
        ),
        
        migrations.AlterField(
            model_name='claimserviceitem',
            name='claim_service',
            field=models.ForeignKey(db_column='ClaimServiceID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='items', to='claim.claimservice'),
        ),
        migrations.AlterField(
            model_name='claimserviceservice',
            name='claim_service',
            field=models.ForeignKey(db_column='claimServiceID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='services', to='claim.claimservice'),
        ),
 
        migrations.AlterField(
            model_name='claimserviceitem',
            name='item',
            field=models.ForeignKey(db_column='ItemID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='service_items', to='medical.item'),
        ),
        migrations.AlterField(
            model_name='claimserviceservice',
            name='service',
            field=models.ForeignKey(db_column='ServiceId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='service_services', to='medical.service'),
        ),
    ]



import logging
import os
import urllib.parse
import uuid
import pathlib
import base64
from urllib.parse import urlparse
from typing import Callable, Dict

import graphene
import importlib
import graphene_django_optimizer
from django.db.models import Count, Case, When, IntegerField, Q, Prefetch

from core.models import MutationLog, Officer
from .apps import ClaimConfig
from claim.validations import validate_claim, get_claim_category, validate_assign_prod_to_claimitems_and_services, \
    process_dedrem, approved_amount
from core import filter_validity, assert_string_length
from core.schema import TinyInt, SmallInt, OpenIMISMutation
from core.gql.gql_mutations import mutation_on_uuids_from_filter
from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError, PermissionDenied, ObjectDoesNotExist
from django.utils.translation import gettext as _
from graphene import InputObjectType
from claim.gql_queries import ClaimGQLType
from claim.models import Claim, Feedback, FeedbackPrompt, ClaimDetail, ClaimItem, ClaimService, ClaimAttachment, \
    ClaimDedRem, GeneralClaimAttachmentType, ClaimAttachmentType,ClaimServiceService
from claim.attachment_strategies import *

from product.models import ProductItemOrService
from medical.models import Item, Service

from claim.utils import process_items_relations, process_services_relations
from claim.services import validate_claim_data as service_validate_claim_data, \
        update_or_create_claim as service_update_or_create_claim, check_unique_claim_code, submit_claim,\
            validate_and_process_dedrem_claim as service_validate_and_process_dedrem_claim,\
            create_feedback_prompt as service_create_feedback_prompt, update_claims_dedrems,\
                set_feedback_prompt_validity_to_to_current_date, set_claims_status
from django.db import transaction
import requests

logger = logging.getLogger(__name__)


class ClaimItemInputType(InputObjectType):
    id = graphene.Int(required=False)
    item_id = graphene.Int(required=True)
    status = TinyInt(required=True)
    qty_provided = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    qty_approved = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_adjusted = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_approved = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_valuated = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    explanation = graphene.String(required=False)
    justification = graphene.String(required=False)
    rejection_reason = SmallInt(required=False)

    validity_from_review = graphene.DateTime(required=False)
    validity_to_review = graphene.DateTime(required=False)
    limitation_value = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    limitation = graphene.String(required=False)
    remunerated_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    deductable_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    exceed_ceiling_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_origin = graphene.String(required=False)
    exceed_ceiling_amount_category = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)

class ClaimSubServiceInputType(InputObjectType):
    id = graphene.Int(required=False)
    sub_service_code = graphene.String(required=True)
    qty_provided = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    qty_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)


class ClaimSubItemInputType(InputObjectType):
    id = graphene.Int(required=False)
    sub_item_code = graphene.String(required=True)
    qty_provided = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    qty_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)

class ClaimServiceInputType(InputObjectType):
    id = graphene.Int(required=False)
    legacy_id = graphene.Int(required=False)
    service_id = graphene.Int(required=True)
    status = TinyInt(required=True)
    qty_provided = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    qty_approved = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_asked = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_adjusted = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_approved = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_valuated = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    explanation = graphene.String(required=False)
    justification = graphene.String(required=False)
    rejection_reason = SmallInt(required=False)
    validity_to = graphene.DateTime(required=False)
    validity_from_review = graphene.DateTime(required=False)
    validity_to_review = graphene.DateTime(required=False)
    audit_user_id_review = graphene.Int(required=False)
    limitation_value = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    limitation = graphene.String(max_length=1, required=False)
    policy_id = graphene.Int(required=False)
    remunerated_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    deductable_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False,
        description="deductable is spelled with a, not deductible")
    exceed_ceiling_amount = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    price_origin = graphene.String(max_length=1, required=False)
    exceed_ceiling_amount_category = graphene.Decimal(
        max_digits=18, decimal_places=2, required=False)
    service_item_set = graphene.List(ClaimSubItemInputType, required=False)
    service_service_set = graphene.List(ClaimSubServiceInputType, required=False)


class FeedbackInputType(InputObjectType):
    id = graphene.Int(required=False, read_only=True)
    care_rendered = graphene.Boolean(required=False)
    payment_asked = graphene.Boolean(required=False)
    drug_prescribed = graphene.Boolean(required=False)
    drug_received = graphene.Boolean(required=False)
    asessment = SmallInt(
        required=False,
        description="Be careful, this field name has a typo")
    officer_id = graphene.Int(required=False)
    feedback_date = graphene.DateTime(required=False)
    validity_from = graphene.DateTime(required=False)
    validity_to = graphene.DateTime(required=False)


class ClaimCodeInputType(graphene.String):

    @staticmethod
    def coerce_string(value):
        assert_string_length(value, ClaimConfig.max_claim_length)
        return value

    serialize = coerce_string
    parse_value = coerce_string

    @staticmethod
    def parse_literal(ast):
        result = graphene.String.parse_literal(ast)
        assert_string_length(result, ClaimConfig.max_claim_length)
        return result


class ClaimGuaranteeIdInputType(graphene.String):

    @staticmethod
    def coerce_string(value):
        assert_string_length(value, 50)
        return value

    serialize = coerce_string
    parse_value = coerce_string

    @staticmethod
    def parse_literal(ast):
        result = graphene.String.parse_literal(ast)
        assert_string_length(result, 50)
        return result


class BaseAttachment:
    id = graphene.String(required=False, read_only=True)
    type = graphene.String(required=False)
    title = graphene.String(required=False)
    date = graphene.Date(required=False)
    filename = graphene.String(required=False)
    mime = graphene.String(required=False)
    general_type = graphene.String(required=False)
    predefined_type = graphene.String(required=False)
    url = graphene.String(required=False)


class BaseAttachmentInputType(BaseAttachment, OpenIMISMutation.Input):
    """
    Claim attachment (without the document), used on its own
    """
    claim_uuid = graphene.String(required=True)


class Attachment(BaseAttachment):
    document = graphene.String(required=False)


class ClaimAttachmentInputType(Attachment, InputObjectType):
    """
    Claim attachment, used nested in claim object
    """
    pass


class AttachmentInputType(Attachment, OpenIMISMutation.Input):
    """
    Claim attachment, used on its own
    """
    claim_uuid = graphene.String(required=True)


class ClaimInputType(OpenIMISMutation.Input):
    id = graphene.Int(required=False, read_only=True)
    uuid = graphene.String(required=False)
    code = ClaimCodeInputType(required=True)
    autogenerate = graphene.Boolean(required=False)
    insuree_id = graphene.Int(required=True)
    date_from = graphene.Date(required=True)
    date_to = graphene.Date(required=False)
    icd_id = graphene.Int(required=True)
    icd_1_id = graphene.Int(required=False)
    icd_2_id = graphene.Int(required=False)
    icd_3_id = graphene.Int(required=False)
    icd_4_id = graphene.Int(required=False)
    review_status = TinyInt(required=False)
    date_claimed = graphene.Date(required=True)
    date_processed = graphene.Date(required=False)
    health_facility_id = graphene.Int(required=True)
    refer_from_id = graphene.Int(required=False)
    refer_to_id = graphene.Int(required=False)
    batch_run_id = graphene.Int(required=False)
    category = graphene.String(max_length=1, required=False)
    visit_type = graphene.String(max_length=1, required=False)
    admin_id = graphene.Int(required=False)
    guarantee_id = ClaimGuaranteeIdInputType(required=False)
    explanation = graphene.String(required=False)
    adjustment = graphene.String(required=False)
    json_ext = graphene.types.json.JSONString(required=False)
    restore = graphene.UUID(required=False)
    feedback_available = graphene.Boolean(default=False)
    feedback_status = TinyInt(required=False)
    feedback = graphene.Field(FeedbackInputType, required=False)
    care_type = graphene.String(required=False)

    items = graphene.List(ClaimItemInputType, required=False)
    services = graphene.List(ClaimServiceInputType, required=False)



class CreateClaimInputType(ClaimInputType):
    attachments = graphene.List(ClaimAttachmentInputType, required=False)


def create_file(date, claim_id, document):
    date_iso = date.isoformat()
    root = ClaimConfig.claim_attachments_root_path
    file_dir = '%s/%s/%s/%s' % (
        date_iso[0:4],
        date_iso[5:7],
        date_iso[8:10],
        claim_id
    )
    file_path = '%s/%s' % (file_dir, uuid.uuid4())
    pathlib.Path('%s/%s' % (root, file_dir)).mkdir(parents=True, exist_ok=True)
    f = open('%s/%s' % (root, file_path), "xb")
    f.write(base64.b64decode(document))
    f.close()
    return file_path


def create_attachment(claim_id, data):
    data["claim_id"] = claim_id
    from core import datetime
    now = datetime.datetime.now()
    general_type = data['general_type']
    data['module'] = 'claim'
    if general_type == GeneralClaimAttachmentType.URL:
        parsed_url = urlparse(data['url'])
        if (ClaimConfig.allowed_domains_attachments and
                not any(domain in parsed_url.path for domain in ClaimConfig.allowed_domains_attachments)):
            raise ValidationError(_("mutation.attachment_url_domain_not_allowed"))
        if data['predefined_type'] in attachment_strategies_dict:
            data['url'] = attachment_strategies_dict[data['predefined_type']].handler(data)
            data['document'] = data['url']
        data['predefined_type'] = ClaimAttachmentType.objects.get(validity_to__isnull=True, claim_general_type="URL",
                                                                  claim_attachment_type=data['predefined_type'])
    elif general_type == GeneralClaimAttachmentType.FILE:
        if ClaimConfig.claim_attachments_root_path:
            # don't use data date as it may be updated by user afterwards!
            data['url'] = create_file(now, claim_id, data.pop('document'))
        data['predefined_type'] = ClaimAttachmentType.objects.get(validity_to__isnull=True, claim_general_type="FILE",
                                                                  claim_attachment_type=data['predefined_type'])
    else:
        raise ValidationError(_("mutation.attachment_general_type_incorrect"))
    data['validity_from'] = now
    ClaimAttachment.objects.create(**data)


def create_attachments(claim_id, attachments):
    for attachment in attachments:
        create_attachment(claim_id, attachment)


def validate_claim_data(data, user):
    return service_validate_claim_data(data, user)

@transaction.atomic
def update_or_create_claim(data, user):
    if "client_mutation_id" in data:
        data.pop('client_mutation_id')
    if "client_mutation_label" in data:
        data.pop('client_mutation_label') 
    return service_update_or_create_claim(data, user)


class CreateClaimMutation(OpenIMISMutation):
    """
    Create a new claim. The claim items and services can all be entered with this call
    """
    _mutation_module = "claim"
    _mutation_class = "CreateClaimMutation"

    class Input(CreateClaimInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            # TODO move this verification to OIMutation
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError(
                    _("mutation.authentication_required"))
            if not user.has_perms(ClaimConfig.gql_mutation_create_claims_perms):
                raise PermissionDenied(_("unauthorized"))
            is_claim_code_autogenerated = data.get("autogenerate", False)
            data['status'] = Claim.STATUS_ENTERED
            from core.utils import TimeUtils
            data['validity_from'] = TimeUtils.now()
            attachments = data.pop('attachments') if 'attachments' in data else None
            claim = update_or_create_claim(data, user)
            if attachments:
                create_attachments(claim.id, attachments)
            if is_claim_code_autogenerated:
                return {"client_mutation_label": f"Create Claim - {claim.code}", "code": f"{claim.code}"}
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_create_claim") % {'code': data['code']},
                'detail': str(exc)}]


class UpdateClaimMutation(OpenIMISMutation):
    """
    Update a claim. The claim items and services can all be updated with this call
    """
    _mutation_module = "claim"
    _mutation_class = "UpdateClaimMutation"

    class Input(ClaimInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            # TODO move this verification to OIMutation
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError(
                    _("mutation.authentication_required"))
            if not user.has_perms(ClaimConfig.gql_mutation_update_claims_perms):
                raise PermissionDenied(_("unauthorized"))
            data['audit_user_id'] = user.id_for_audit
            update_or_create_claim(data, user)
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_update_claim") % {'code': data['code']},
                'detail': str(exc)}]


class CreateAttachmentMutation(OpenIMISMutation):
    _mutation_module = "claim"
    _mutation_class = "AddClaimAttachmentMutation"

    class Input(AttachmentInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        claim = None
        try:
            if user.is_anonymous or not user.has_perms(ClaimConfig.gql_mutation_update_claims_perms):
                raise PermissionDenied(_("unauthorized"))
            if "client_mutation_id" in data:
                data.pop('client_mutation_id')
            if "client_mutation_label" in data:
                data.pop('client_mutation_label')
            claim_uuid = data.pop("claim_uuid")
            queryset = Claim.objects.filter(*filter_validity())
            if settings.ROW_SECURITY:
                from location.schema import LocationManager
                queryset = LocationManager().build_user_location_filter_query( user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])           
            claim = queryset.filter(uuid=claim_uuid).first()
            if not claim:
                raise PermissionDenied(_("unauthorized"))
            create_attachment(claim.id, data)
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_attach_document") % {'code': claim.code if claim else None},
                'detail': str(exc)}]


class UpdateAttachmentMutation(OpenIMISMutation):
    _mutation_module = "claim"
    _mutation_class = "UpdateAttachmentMutation"

    class Input(BaseAttachmentInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if not user.has_perms(ClaimConfig.gql_mutation_update_claims_perms):
                raise PermissionDenied(_("unauthorized"))
            queryset = ClaimAttachment.objects.filter(*filter_validity())
            if settings.ROW_SECURITY:
                from location.schema import LocationManager
                queryset = LocationManager().build_user_location_filter_query( user._u, prefix='claim__health_facility__location', queryset = queryset.select_related("claim"), loc_types=['D'])

            attachment = queryset \
                .filter(id=data['id']) \
                .first()
            if not attachment:
                raise PermissionDenied(_("unauthorized"))
            general_type = data['general_type']
            data['module'] = 'claim'
            from core import datetime
            now = datetime.datetime.now()
            if general_type == GeneralClaimAttachmentType.URL:
                parsed_url = urlparse(data['url'])
                if (ClaimConfig.allowed_domains_attachments and
                        not any(domain in parsed_url.path for domain in ClaimConfig.allowed_domains_attachments)):
                    raise ValidationError(_("mutation.attachment_url_domain_not_allowed"))
                if data['predefined_type'] in attachment_strategies_dict:
                    data['url'] = attachment_strategies_dict[data['predefined_type']].handler(data)
                    data['document'] = data['url']
                data['predefined_type'] = ClaimAttachmentType.objects.get(validity_to__isnull=True,
                                                                          claim_general_type="URL",
                                                                          claim_attachment_type=data['predefined_type'])
            elif general_type == GeneralClaimAttachmentType.FILE:
                if ClaimConfig.claim_attachments_root_path:
                    # don't use data date as it may be updated by user afterwards!
                    data['url'] = create_file(now, claim_id, data.pop('document'))
                data['predefined_type'] = ClaimAttachmentType.objects.get(validity_to__isnull=True,
                                                                          claim_general_type="FILE",
                                                                          claim_attachment_type=data['predefined_type'])
            attachment.save_history()
            data['audit_user_id'] = user.id_for_audit
            [setattr(attachment, key, data[key]) for key in data]
            attachment.save()
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_update_claim_attachment") % {
                    'code': attachment.claim.code,
                    'filename': attachment.filename
                },
                'detail': str(exc)}]


class DeleteAttachmentMutation(OpenIMISMutation):
    _mutation_module = "claim"
    _mutation_class = "DeleteClaimAttachmentMutation"

    class Input(OpenIMISMutation.Input):
        id = graphene.String()

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if not user.has_perms(ClaimConfig.gql_mutation_update_claims_perms):
                raise PermissionDenied(_("unauthorized"))
            queryset = ClaimAttachment.objects.filter(*filter_validity())
            if settings.ROW_SECURITY:
                from location.schema import LocationManager
                queryset = LocationManager().build_user_location_filter_query( user._u, prefix='health_facility__location', queryset = queryset, loc_types=['D'])     
            attachment = queryset \
                .filter(id=data['id']) \
                .first()
            if not attachment:
                raise PermissionDenied(_("unauthorized"))
            attachment.delete_history()
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_delete_claim_attachment") % {
                    'code': attachment.claim.code,
                    'filename': attachment.filename
                },
                'detail': str(exc)}]


class ClaimSubmissionStatsMixin:
    @classmethod
    def _generate_claim_submission_stats(cls, uuids):
        claims_query = Claim.objects.filter(uuid__in=list(uuids))
        claim_item_query = ClaimItem.objects.filter(claim__in=claims_query)
        claim_service_query = ClaimService.objects.filter(claim__in=claims_query)
        claim_stats = claims_query.aggregate(
            submitted=Count('uuid', output_field=IntegerField()),
            checked=Count(Case(When(status=4, then=1), output_field=IntegerField())),
            processed=Count(Case(When(status=8, then=1), output_field=IntegerField())),
            valuated=Count(Case(When(status=16, then=1), output_field=IntegerField())),
            rejected=Count(Case(When(status=1, then=1), output_field=IntegerField())),
        )
        item_stats = claim_item_query.aggregate(
            items_passed=Count(Case(When(status=1, then=1), output_field=IntegerField())),
            items_rejected=Count(Case(When(status=2, then=1), output_field=IntegerField())),
        )
        service_stats = claim_service_query.aggregate(
            services_passed=Count(Case(When(status=1, then=1), output_field=IntegerField())),
            services_rejected=Count(Case(When(status=2, then=1), output_field=IntegerField())),
        )

        return {**claim_stats, **item_stats, **service_stats}

    @classmethod
    def _parse_submission_stats(cls, claim_submission_stats):
        return claim_submission_stats

    @classmethod
    def add_submission_stats_to_mutation_log(cls, client_mutation_id, uuids):
        mutation_log = MutationLog.objects.filter(client_mutation_id=client_mutation_id).first()
        claim_submission_stats = cls._generate_claim_submission_stats(uuids)
        parsed_stats = cls._parse_submission_stats(claim_submission_stats)
        if isinstance(mutation_log.json_ext, dict):
            mutation_log.json_ext["claim_stats"] = parsed_stats
        else:
            mutation_log.json_ext = {"claim_stats": parsed_stats}
        mutation_log.save()


class SubmitClaimsMutation(OpenIMISMutation, ClaimSubmissionStatsMixin):
    """
    Submit one or several claims.
    """
    __filter_handlers = {
        'services': 'services__service__code__in',
        'items': 'items__item__code__in'
    }
    _mutation_module = "claim"
    _mutation_class = "SubmitClaimsMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)
        additional_filters = graphene.String()

    @classmethod
    def _parse_submission_stats(cls, claim_submission_stats):
        return {
            "submitted": claim_submission_stats["submitted"],
            "checked": claim_submission_stats["checked"],
            "rejected": claim_submission_stats["rejected"],
            "items_passed": claim_submission_stats["items_passed"],
            "items_rejected": claim_submission_stats["items_rejected"],
            "services_passed": claim_submission_stats["services_passed"],
            "services_rejected": claim_submission_stats["services_rejected"],
            "header": "Claims submitted",
            # failed
        }

    @classmethod
    @mutation_on_uuids_from_filter(Claim, ClaimGQLType, 'additional_filters', __filter_handlers)
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_submit_claims_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = []
        uuids = data.get("uuids", [])
        client_mutation_id = data.get("client_mutation_id", None)
        c_errors = []
        claims = Claim.objects \
                .filter(uuid__in=uuids,
                        validity_to__isnull=True) \
            .prefetch_related(Prefetch('items', queryset=ClaimItem.objects.filter(
                *filter_validity(), 
                Q(Q(rejection_reason=0) | Q(rejection_reason__isnull=True))))) \
            .prefetch_related(Prefetch('services', queryset=ClaimService.objects.filter(
                *filter_validity(),
                Q(Q(rejection_reason=0) | Q(rejection_reason__isnull=True))))) 
        remaining_uuid = list(map(str.upper,uuids))
        for claim in claims:
            remaining_uuid.remove(claim.uuid.upper())
            c_errors += submit_claim(claim, user)
            if c_errors:
                errors.append({
                    'title': claim.code,
                    'list': c_errors
                })
        if len(remaining_uuid):
            c_errors.append( {'code': REJECTION_REASON_INVALID_CLAIM,
                            'message': _("claim.validation.claim_uuid_not_found") + ','.join(remaining_uuid) })
        if len(errors) == 1:
            errors = errors[0]['list']
        cls.add_submission_stats_to_mutation_log(client_mutation_id, uuids)
        logger.debug("SubmitClaimsMutation: claim done, errors: %s", len(errors))
        return errors



def create_feedback_prompt(claim_uuid, user):
    current_claim = Claim.objects.get(uuid=claim_uuid)
    return service_create_feedback_prompt(current_claim, user)
    


class SelectClaimsForFeedbackMutation(OpenIMISMutation):
    """
    Select one or several claims for feedback.
    """
    _mutation_module = "claim"
    _mutation_class = "SelectClaimsForFeedbackMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_select_claim_feedback_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'feedback_status', Claim.FEEDBACK_SELECTED, user=user)


class BypassClaimsFeedbackMutation(OpenIMISMutation):
    """
    Bypass feedback for one or several claims
    """
    _mutation_module = "claim"
    _mutation_class = "BypassClaimsFeedbackMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_bypass_claim_feedback_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'feedback_status', Claim.FEEDBACK_BYPASSED)


class SkipClaimsFeedbackMutation(OpenIMISMutation):
    """
    Skip feedback for one or several claims
    Skip indicates that the claim is not selected for feedback
    """
    _mutation_module = "claim"
    _mutation_class = "SkipClaimsFeedbackMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_skip_claim_feedback_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'feedback_status', Claim.FEEDBACK_NOT_SELECTED)


class DeliverClaimFeedbackMutation(OpenIMISMutation):
    """
    Deliver feedback of a claim
    """
    _mutation_module = "claim"
    _mutation_class = "DeliverClaimFeedbackMutation"

    class Input(OpenIMISMutation.Input):
        claim_uuid = graphene.String(required=False, read_only=True)
        feedback = graphene.Field(FeedbackInputType, required=True)

    @classmethod
    def async_mutate(cls, user, **data):
        claim = None
        try:
            if not user.has_perms(ClaimConfig.gql_mutation_deliver_claim_feedback_perms):
                raise PermissionDenied(_("unauthorized"))
            claim = Claim.objects.select_related('feedback').get(
                uuid=data['claim_uuid'],
                validity_to__isnull=True)
            prev_feedback = claim.feedback
            prev_claim_id = claim.save_history()
            if prev_feedback:
                prev_feedback.claim_id = prev_claim_id
                prev_feedback.save()
            feedback = data['feedback']
            from core.utils import TimeUtils
            feedback['validity_from'] = TimeUtils.now()
            feedback['audit_user_id'] = user.id_for_audit
            # The legacy model has a Foreign key on both sides of this one-to-one relationship
            f, created = Feedback.objects.update_or_create(
                claim=claim,
                defaults=feedback
            )
            claim.feedback = f
            claim.feedback_status = Claim.FEEDBACK_DELIVERED
            claim.feedback_available = True
            claim.save()
            set_feedback_prompt_validity_to_to_current_date(claim.uuid)
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_update_claim") % {'code': claim.code if claim else None},
                'detail': str(exc)}]


class SelectClaimsForReviewMutation(OpenIMISMutation):
    """
    Select one or several claims for review.
    """
    _mutation_module = "claim"
    _mutation_class = "SelectClaimsForReviewMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_select_claim_review_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'review_status', Claim.REVIEW_SELECTED)


class BypassClaimsReviewMutation(OpenIMISMutation):
    """
    Bypass review for one or several claims
    Bypass indicates that review of a previously selected claim won't be delivered
    """
    _mutation_module = "claim"
    _mutation_class = "BypassClaimsReviewMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_bypass_claim_review_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'review_status', Claim.REVIEW_BYPASSED)


class DeliverClaimsReviewMutation(OpenIMISMutation):
    """
    Mark claim review as delivered for one or several claims
    """
    _mutation_module = "claim"
    _mutation_class = "DeliverClaimsReviewMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        logger.error("SaveClaimReviewMutation")
        if not user.has_perms(ClaimConfig.gql_mutation_deliver_claim_review_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = set_claims_status(data['uuids'], 'review_status', Claim.REVIEW_DELIVERED,
                                   {'audit_user_id_review': user.id_for_audit})
        # OMT-208 update the dedrem for the reviewed claims
        errors += update_claims_dedrems(data["uuids"], user)

        return errors


class SkipClaimsReviewMutation(OpenIMISMutation):
    """
    Skip review for one or several claims
    Skip indicates that the claim is not selected for review
    """
    _mutation_module = "claim"
    _mutation_class = "SkipClaimsReviewMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_skip_claim_review_perms):
            raise PermissionDenied(_("unauthorized"))
        return set_claims_status(data['uuids'], 'review_status', Claim.REVIEW_NOT_SELECTED)


class SaveClaimReviewMutation(OpenIMISMutation):
    """
    Save the review of a claim (items and services)
    """
    _mutation_module = "claim"
    _mutation_class = "SaveClaimReviewMutation"

    class Input(OpenIMISMutation.Input):
        claim_uuid = graphene.String(required=False, read_only=True)
        adjustment = graphene.String(required=False)
        items = graphene.List(ClaimItemInputType, required=False)
        services = graphene.List(ClaimServiceInputType, required=False)

    @classmethod
    def async_mutate(cls, user, **data):
        claim = None
        try:
            if not user.has_perms(ClaimConfig.gql_mutation_deliver_claim_review_perms):
                raise PermissionDenied(_("unauthorized"))
            claim = Claim.objects.get(uuid=data['claim_uuid'],
                                      validity_to__isnull=True)
            if claim is None:
                return [{'message': _(
                    "claim.validation.id_does_not_exist") % {'id': data['claim_uuid']}}]
            claim.save_history()
            claim.adjustment = data.get('adjustment', None)
            items = data.pop('items') if 'items' in data else []
            all_rejected = True
            for item in items:
                item_id = item.pop('id')
                claim.items.filter(id=item_id).update(**item)
                if item['status'] == ClaimItem.STATUS_PASSED:
                    all_rejected = False
            services = data.pop('services') if 'services' in data else []
            claimed = 0
            claim_service_elements = []
            for service in services:
                service_id = service.pop('id')
                service_linked = service.pop('service_service_set', [])
                logger.debug("service_linked ", service_linked)
                service_service_set = service.pop('service_service_set', [])
                logger.debug("service_service_set ", service_service_set)
                claim.services.filter(id=service_id).update(**service)
                if ClaimConfig.native_code_for_services == False:
                    for claim_service_service in service_service_set:
                        claim_service_code = claim_service_service.pop('subServiceCode')
                        claim_service = claim.services.filter(id=service_id).first()
                        if claim_service:
                            service_element = Service.objects.filter(*filter_validity(), code=claim_service_code).first()
                            if service_element:
                                claim_service_to_update = claim_service.services.filter(service=service_element.id)
                                logger.debug("claim_service_to_update ", claim_service_to_update)
                                if claim_service_to_update:
                                    qty_asked = claim_service_service.pop('qty_asked', 0)
                                    price_asked = claim_service_service.pop('price_asked', 0)
                                    claim_service_service['qty_displayed'] = qty_asked
                                    price = qty_asked * price_asked
                                    claimed += price
                                    claim_service_to_update.update(**claim_service_service)
                            claim_service_elements.append(claim_service)
                    for claim_service_item in service_linked:
                        claim_item_code = claim_service_item.pop('subItemCode')
                        claim_service = claim.services.filter(id=service_id).first()
                        if claim_service:
                            item_element = Item.objects.filter(*filter_validity(), code=claim_item_code).first()
                            if item_element:
                                claim_item_to_update = claim_service.items.filter(item=item_element.id)
                                logger.debug("claim_item_to_update ", claim_item_to_update)
                                if claim_item_to_update:
                                    qty_asked = claim_service_item.pop('qty_asked', 0)
                                    price_asked = claim_service_item.pop('price_asked', 0)
                                    claim_service_item['qty_displayed'] = qty_asked
                                    price = qty_asked * price_asked
                                    claimed += price
                                    claim_item_to_update.update(**claim_service_item)

                if service['status'] == ClaimService.STATUS_PASSED:
                    all_rejected = False
            claim.approved = approved_amount(claim)
            if ClaimConfig.native_code_for_services == False:
                claim.claimed = claimed
                for claimservice in claim_service_elements:
                    setattr(claimservice, 'price_adjusted', claimed)
            claim.audit_user_id_review = user.id_for_audit
            if all_rejected:
                claim.status = Claim.STATUS_REJECTED
            claim.save()
            return None
        except Exception as exc:
            return [{
                'message': _("claim.mutation.failed_to_update_claim") % {'code': claim.code if claim else None},
                'detail': str(exc)}]


class ProcessClaimsMutation(OpenIMISMutation, ClaimSubmissionStatsMixin):
    """
    Process one or several claims.
    """
    _mutation_module = "claim"
    _mutation_class = "ProcessClaimsMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def _parse_submission_stats(cls, claim_submission_stats):
        return {
            "submitted": claim_submission_stats["submitted"],
            "processed": claim_submission_stats["processed"],
            "valuated": claim_submission_stats["valuated"],
            "rejected": claim_submission_stats["rejected"],
            "header": "Submitted to process",
            # failed
        }

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_process_claims_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = []
        uuids = data.get("uuids", None)
        client_mutation_id = data.get("client_mutation_id", None)
        claims = Claim.objects \
                .filter(uuid__in=uuids) \
                .prefetch_related(Prefetch('items', queryset=ClaimItem.objects.filter(*filter_validity())))\
                .prefetch_related(Prefetch('services', queryset=ClaimService.objects.filter(*filter_validity())))
        remaining_uuid = list(map(str.upper,uuids))
        for claim in claims:
            remaining_uuid.remove(claim.uuid.upper())
            

            logger.debug("ProcessClaimsMutation: processing %s", claim.uuid)
            c_errors = []
     
            claim.save_history()
            claim.audit_user_id_process = user.id_for_audit
            logger.debug("ProcessClaimsMutation: validating claim %s", claim.uuid)
            c_errors += validate_and_process_dedrem_claim(claim, user, True)

            logger.debug("ProcessClaimsMutation: claim %s set processed or valuated", claim.uuid)
            if c_errors:
                errors.append({
                    'title': claim.code,
                    'list': c_errors
                })
                
        if len(remaining_uuid):
                errors += {
                    'title': _('error'),
                    'list': [{'message': _(
                        "claim.validation.id_does_not_exist") % {'id': ','.join(remaining_uuid)}}]
                }
        if len(errors) == 1:
            errors = errors[0]['list']
        cls.add_submission_stats_to_mutation_log(client_mutation_id, uuids)
        logger.debug("ProcessClaimsMutation: claims %s done, errors: %s", data["uuids"], len(errors))
        return errors


class DeleteClaimsMutation(OpenIMISMutation):
    """
    Mark one or several claims as Deleted (validity_to)
    """

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    _mutation_module = "claim"
    _mutation_class = "DeleteClaimsMutation"

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ClaimConfig.gql_mutation_delete_claims_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = []
        for claim_uuid in data["uuids"]:
            claim = Claim.objects \
                .filter(uuid=claim_uuid) \
                .prefetch_related(Prefetch('items', queryset=ClaimItem.objects.filter(*filter_validity())))\
                .prefetch_related(Prefetch('services', queryset=ClaimService.objects.filter(*filter_validity())))\
                .first()
            if claim is None:
                errors += {
                    'title': claim_uuid,
                    'list': [{'message': _(
                        "claim.validation.id_does_not_exist") % {'id': claim_uuid}}]
                }
                continue
            errors += set_claim_deleted(claim)
        if len(errors) == 1:
            errors = errors[0]['list']
        return errors





def set_claim_deleted(claim):
    try:
        claim.delete_history()
        return []
    except Exception as exc:
        return {
            'title': claim.code,
            'list': [{
                'message': _("claim.mutation.failed_to_change_status_of_claim") % {'code': claim.code},
                'detail': claim.uuid}]
        }




def validate_and_process_dedrem_claim(claim, user, is_process):
    return service_validate_and_process_dedrem_claim(claim, user, is_process)



# Generated by Django 3.2.19 on 2023-12-05 14:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0025_mutationlog_json_ext'),
        ('claim', '0024_add_claim_restore_perms'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feedbackprompt',
            name='officer_id',
            field=models.ForeignKey(blank=True, db_column='OfficerID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='core.officer'),
        ),
        migrations.AlterField(
            model_name='feedbackprompt',
            name='claim_id',
            field=models.ForeignKey(blank=True, db_column='ClaimID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='claim.claim'),
        ),
        migrations.RenameField(
            model_name='feedbackprompt',
            old_name='officer_id',
            new_name='officer'
        ),
        migrations.RenameField(
            model_name='feedbackprompt',
            old_name='claim_id',
            new_name='claim'
        )
        
    ]



# Generated by Django 2.1.14 on 2020-03-18 10:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0005_claimmutation'),
    ]

    operations = [
        migrations.AddField(
            model_name='claimattachment',
            name='url',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.test import TestCase
from unittest import mock
from location.test_helpers import create_test_location, create_test_health_facility,create_test_village
from insuree.test_helpers import create_test_insuree
from claim.test_helpers import create_test_claim_admin, create_test_claim
from claim.models import Claim, ClaimItem, ClaimService,ClaimDetail
from medical.models import  Diagnosis, Item, Service
from medical.test_helpers import create_test_item, create_test_service

from core.services import create_or_update_interactive_user, create_or_update_core_user
import datetime
from claim.services import *
import core
from medical.test_helpers import create_test_service, create_test_item
from claim.utils import service_create_hook, calcul_amount_service, service_update_hook
from claim.test_helpers import create_test_claim
from claim.models import ClaimServiceItem, ClaimServiceService

class ClaimSubmitServiceTestCase(TestCase):
    test_hf = None

    test_insuree =None
    test_claim_admin = None
    test_icd = None
    test_claim = None
    test_claim_item = None
    test_claim_service = None
    test_region = None
    test_district = None
    test_village = None
    test_ward = None
    

    @classmethod
    def setUpTestData(cls):
        if cls.test_region is None:
            cls.test_village  =create_test_village( )
            cls.test_ward =cls.test_village.parent
            cls.test_region =cls.test_village.parent.parent.parent
            cls.test_district = cls.test_village.parent.parent

        cls.test_hf=create_test_health_facility("1", cls.test_district.id, valid=True)
        props = dict(
            last_name="name",
            other_names="surname",
            dob=core.datetime.date(2000, 1, 13),
            chf_id="884930485",
        )
        family_props = dict(
            location=cls.test_village,
        )
        cls.test_insuree= create_test_insuree(is_head=True, custom_props=props, family_custom_props=family_props)
        cls.test_claim_admin= create_test_claim_admin()
        cls.test_icd = Diagnosis(code='ICD00I', name='diag test', audit_user_id=-1)
        cls.test_icd.save()
        cls.test_claim = Claim.objects.create(
            date_claimed=core.datetime.date(2020, 1, 9),
            code="code_ABVC",
            icd=cls.test_icd,
            claimed=2000,
            date_from=core.datetime.date(2020, 1, 13),
            admin=cls.test_claim_admin,
            insuree=cls.test_insuree,
            health_facility=cls.test_hf,
            status=Claim.STATUS_ENTERED,
            audit_user_id=-1
        )
        
        cls.test_claim_item = ClaimItem.objects.create(
            claim = cls.test_claim,
            item =create_test_item(
                'D',
                custom_props={"code": "cCode", "price" :1000}
            ),
            price_asked = 1000,
            qty_provided=1,
            audit_user_id=-1,
            status=ClaimDetail.STATUS_PASSED,
            availability=True
        )
        cls.test_claim_service = ClaimService.objects.create(
            claim = cls.test_claim,
            service = create_test_service(
                'D',
                custom_props={"code": "sCode", "price" :1000}
            ),
            price_asked = 1000,
            qty_provided=1,
            audit_user_id=-1,
            status=ClaimDetail.STATUS_PASSED
        )

    def test_minimal_item_claim_submit_xml(self):
        items = [
            ClaimItemSubmit(code='aa', quantity=2),
        ]
        item = "<Item><ItemCode>aa</ItemCode><ItemQuantity>2</ItemQuantity></Item>"

        claim = ClaimSubmit(
            date=core.datetime.date(2020, 1, 9),
            code="code_ABVC",
            icd_code=self.test_icd.code,
            total=334,
            start_date=core.datetime.date(2020, 1, 13),
            claim_admin_code=self.test_claim_admin.code,
            insuree_chf_id=self.test_insuree.chf_id,
            health_facility_code=self.test_hf.code,
            item_submits=items,
        )
        details = "<Details>"
        details = details + "<ClaimDate>09/01/2020</ClaimDate>"
        details = details + f"<HFCode>{self.test_hf.code}</HFCode>"
        details = details + f"<ClaimAdmin>{self.test_claim_admin.code}</ClaimAdmin>"
        details = details + "<ClaimCode>code_ABVC</ClaimCode>"
        details = details + f"<CHFID>{self.test_insuree.chf_id}</CHFID>"
        details = details + "<StartDate>13/01/2020</StartDate>"
        details = details + f"<ICDCode>{self.test_icd.code}</ICDCode>"
        details = details + "<Total>334</Total>"
        details = details + "</Details>"
        expected = "<Claim>%s<Items>%s</Items></Claim>" % (details, item)
        self.assertEquals(expected, claim.to_xml())

    def test_minimal_service_claim_submit_xml(self):
        services = [
            ClaimServiceSubmit(code='aa', quantity=2),
        ]
        service = "<Service><ServiceCode>aa</ServiceCode><ServiceQuantity>2</ServiceQuantity></Service>"

        claim = ClaimSubmit(
            date=core.datetime.date(2020, 1, 9),
            code="code_ABVC",
            icd_code=self.test_icd.code,
            total=334,
            start_date=core.datetime.date(2020, 1, 13),
            claim_admin_code=self.test_claim_admin.code,
            insuree_chf_id=self.test_insuree.chf_id,
            health_facility_code=self.test_hf.code,
            service_submits=services,
        )
        details = "<Details>"
        details = details + "<ClaimDate>09/01/2020</ClaimDate>"
        details = details + f"<HFCode>{self.test_hf.code}</HFCode>"
        details = details + f"<ClaimAdmin>{self.test_claim_admin.code}</ClaimAdmin>"
        details = details + "<ClaimCode>code_ABVC</ClaimCode>"
        details = details + f"<CHFID>{self.test_insuree.chf_id}</CHFID>"
        details = details + "<StartDate>13/01/2020</StartDate>"
        details = details + f"<ICDCode>{self.test_icd.code}</ICDCode>"
        details = details + "<Total>334</Total>"
        details = details + "</Details>"
        expected = "<Claim>%s<Services>%s</Services></Claim>" % (
            details, service)
        self.assertEquals(expected, claim.to_xml())

    def test_extended_claim_submit_xml(self):
        items = [
            ClaimItemSubmit(code='aa', quantity=2),
            ClaimItemSubmit(code='bb', quantity=1, price=12.3),
        ]
        item_a = "<Item><ItemCode>aa</ItemCode><ItemQuantity>2</ItemQuantity></Item>"
        item_b = "<Item><ItemCode>bb</ItemCode><ItemPrice>12.3</ItemPrice><ItemQuantity>1</ItemQuantity></Item>"

        services = [
            ClaimServiceSubmit(code='aa-serv', quantity=2),
            ClaimServiceSubmit(code='a<a\'-serv', quantity=1, price=35),
        ]
        service_a = "<Service><ServiceCode>aa-serv</ServiceCode><ServiceQuantity>2</ServiceQuantity></Service>"
        service_b = "<Service><ServiceCode>a&lt;a\'-serv</ServiceCode><ServicePrice>35</ServicePrice><ServiceQuantity>1</ServiceQuantity></Service>"

        claim = ClaimSubmit(
            date=core.datetime.date(2020, 1, 9),
            code="code_ABVC",
            icd_code=self.test_icd.code,
            total=334,
            start_date=core.datetime.date(2020, 1, 13),
            claim_admin_code=self.test_claim_admin.code,
            insuree_chf_id=self.test_insuree.chf_id,
            health_facility_code=self.test_hf.code,
            item_submits=items,
            service_submits=services
        )
          
        details = "<Details>"
        details = details + "<ClaimDate>09/01/2020</ClaimDate>"
        details = details + f"<HFCode>{self.test_hf.code}</HFCode>"
        details = details + f"<ClaimAdmin>{self.test_claim_admin.code}</ClaimAdmin>"
        details = details + "<ClaimCode>code_ABVC</ClaimCode>"
        details = details + f"<CHFID>{self.test_insuree.chf_id}</CHFID>"        
        details = details + "<StartDate>13/01/2020</StartDate>"
        details = details + f"<ICDCode>{self.test_icd.code}</ICDCode>"
        details = details + "<Total>334</Total>"
        details = details + "</Details>"
        expected = "<Claim>%s" % details
        expected = expected + "<Items>%s%s</Items>" % (item_a, item_b)
        expected = expected + \
                   "<Services>%s%s</Services>" % (service_a, service_b)
        expected = expected + "</Claim>"
        self.assertEquals(expected, claim.to_xml())

    @mock.patch('django.db.connections')
    def test_claim_submit_error(self, mock_connections):
        if connection.vendor != 'microsoft':
            self.skipTest("This test can only be executed for MSSQL database")
        with mock.patch("claim.services.ClaimSubmitService.hf_scope_check") as mock_security:
            mock_security.return_value = None
            query_result = [2]
            mock_connections.__getitem__.return_value.cursor.return_value \
                .__enter__.return_value.fetchone.return_value = query_result
            mock_user = mock.Mock(is_anonymous=False)
            mock_user.has_perm = mock.MagicMock(return_value=True)
            claim = ClaimSubmit(
                date=core.datetime.date(2020, 1, 9),
                code="code_ABVC",
                icd_code=self.test_icd.code,
                total=334,
                start_date=core.datetime.date(2020, 1, 13),
                claim_admin_code=self.test_claim_admin.code,
                insuree_chf_id=self.test_insuree.chf_id,
                health_facility_code=self.test_hf.code,
            )
            service = ClaimSubmitService(user=mock_user)
            with self.assertRaises(ClaimSubmitError) as cm:
                service.submit(claim)
            self.assertNotEqual(cm.exception.code, 0)

    @mock.patch('django.db.connections')
    def test_claim_submit_allgood_xml(self, mock_connections):
        if connection.vendor != 'microsoft':
            self.skipTest("This test can only be executed for MSSQL database")
        with mock.patch("django.db.backends.utils.CursorWrapper") as mock_cursor:
            # required for all modules tests
            mock_cursor.return_value.description = None
            # required for claim module tests
            mock_cursor.return_value.__enter__.return_value.description = None
            with mock.patch("claim.services.ClaimSubmitService.hf_scope_check") as mock_security:
                mock_security.return_value = None
                mock_user = mock.Mock(is_anonymous=False)
                mock_user.has_perm = mock.MagicMock(return_value=True)
                claim = ClaimSubmit(
                    date=core.datetime.date(2020, 1, 9),
                    code="code_ABVCD",
                    icd_code=self.test_icd.code,
                    total=334,
                    start_date=core.datetime.date(2020, 1, 13),
                    claim_admin_code=self.test_claim_admin.code,
                    insuree_chf_id=self.test_insuree.chf_id,
                    health_facility_code=self.test_hf.code,
                )
                service = ClaimSubmitService(user=mock_user)
                service.submit(claim)  # doesn't raise an error

    @mock.patch("claim.services.ClaimSubmitService._validate_user_hf")
    @mock.patch("claim.services.ClaimCreateService._validate_user_hf")
    def test_claim_enter_and_submit(self, check_hf_submit, check_hf_enter):
        check_hf_submit.return_value, check_hf_enter.return_value = True, True
        mock_user = mock.Mock(is_anonymous=False)
        mock_user.has_perm = mock.MagicMock(return_value=True)
        mock_user.id_for_audit = -1

        claim = self._get_test_dict(code='e_n_s')
        service = ClaimSubmitService(user=mock_user)
        submitted_claim = service.enter_and_submit(claim, False)
        expected_claimed = 1000 + 1000  # 2 provisions, both qty = 1, price asked == 1000

        self.assertEqual(submitted_claim.status, Claim.STATUS_CHECKED)
        self.assertEqual(submitted_claim.approved, expected_claimed)
        self.assertEqual(submitted_claim.claimed, expected_claimed)
        self.assertEqual(submitted_claim.health_facility.id, self.test_hf.id)
        self.assertEqual(submitted_claim.items.all().count(), 1)
        self.assertEqual(submitted_claim.services.all().count(), 1)
        self.assertEqual(submitted_claim.audit_user_id, -1)
        self.assertTrue(submitted_claim.id is not None)

    @mock.patch("claim.services.ClaimSubmitService._validate_user_hf")
    @mock.patch("claim.services.ClaimCreateService._validate_user_hf")
    def test_claim_enter_duplicate_exception(self, check_hf_submit, check_hf_enter):
        check_hf_submit.return_value, check_hf_enter.return_value = True, True
        mock_user = mock.Mock(is_anonymous=False)
        mock_user.has_perm = mock.MagicMock(return_value=True)
        mock_user.id_for_audit = -1

        claim = self._get_test_dict(code='dup')
        service = ClaimSubmitService(user=mock_user)

        service.enter_and_submit(claim, False)

        with self.assertRaises(ValidationError):
            service.enter_and_submit(claim, False)

    def test_service_create_hook(self):
        item = create_test_item("D", custom_props={})
        service = create_test_service("V")
        claim = create_test_claim()
        service_items_dict = {
            "qty_provided": 7, "price_asked": 11, "service_id": 23,
            "status": 1, "validity_from": "2019-06-01", "validity_to": None, "audit_user_id": -1,
            "service_item_set": [{"sub_item_code": item.code, "qty_asked":1, "qty_provided": 7, "price_asked": 11}],
            "service_service_set": [{"sub_service_code": service.code, "qty_asked":2, "qty_provided": 3, "price_asked": 20}]
        }
        service_create_hook(claim.id, service_items_dict)
        claim_service_item = ClaimServiceItem.objects.filter(item=item.id)
        self.assertTrue(len(claim_service_item) > 0)
        claim_service_item = claim_service_item.first()
        self.assertEquals(claim_service_item.qty_displayed, 1)
        self.assertEquals(claim_service_item.qty_provided, 7)
        self.assertEquals(claim_service_item.price_asked, 11)

        claim_service_service = ClaimServiceService.objects.filter(service=service.id)
        self.assertTrue(len(claim_service_service) > 0)
        claim_service_service = claim_service_service.first()
        self.assertEquals(claim_service_service.qty_displayed, 2)
        self.assertEquals(claim_service_service.qty_provided, 3)
        self.assertEquals(claim_service_service.price_asked, 20)

        service_items_dict["service_item_set"] = [{"qty_asked": 90, "sub_item_code": item.code}]
        service_items_dict["service_service_set"] = [{"qty_asked": 100, "sub_service_code": service.code}]
            
        service_update_hook(claim.id, service_items_dict)
        claim_service_service.refresh_from_db()
        self.assertEqual(100, claim_service_service.qty_displayed)
        
        claim_service_item.refresh_from_db()
        self.assertEqual(90, claim_service_item.qty_displayed)
    
    def test_calcul_amount_service(self):
        item = create_test_item("D", custom_props={})
        service = create_test_service("V")
        service_items_dict = {
            "qty_provided": 5, "price_asked": 50, "service_id": 23,
            "status": 1, "validity_from": "2019-06-01", "validity_to": None, "audit_user_id": -1,
            "service_item_set": [{"sub_item_code": item.code, "qty_asked":1, "qty_provided": 7, "price_asked": 11}],
            "service_service_set": [{"sub_service_code": service.code, "qty_asked":2, "qty_provided": 3, "price_asked": 20}]
        }
        result = calcul_amount_service(service_items_dict)
        self.assertEqual(result, 51)

    def _get_test_dict(self, code=None):
        return {
            "health_facility_id": self.test_claim.health_facility_id, 
            "icd_id": self.test_icd.id, 
            "date_from": self.test_claim.date_from, 
            "code": self.test_claim.code if code is None else code,
            "date_claimed": self.test_claim.date_claimed, 
            "date_to": self.test_claim.date_to,
            "audit_user_id": self.test_claim.audit_user_id, 
            "insuree_id": self.test_claim.insuree_id, 
            "status": self.test_claim.status, 
            "validity_from": self.test_claim.validity_from,
            "items": [{
                "qty_provided": self.test_claim_item.qty_provided, 
                "price_asked": self.test_claim_item.price_asked, 
                "item_id": self.test_claim_item.item_id, 
                "status": self.test_claim_item.status, 
                "availability": self.test_claim_item.availability,
                "validity_from": self.test_claim_item.validity_from, 
                "validity_to": self.test_claim_item.validity_to, 
                "audit_user_id": self.test_claim_item.audit_user_id
            }],
            "services": [{
                "qty_provided": self.test_claim_service.qty_provided, 
                "price_asked": self.test_claim_service.price_asked, 
                "service_id": self.test_claim_service.service_id, 
                "status": self.test_claim_service.status, 
                "validity_from": self.test_claim_service.validity_from, 
                "validity_to": self.test_claim_service.validity_to, 
                "audit_user_id": self.test_claim_service.audit_user_id
            }]
        }



import math
from claim.models import Claim, ClaimItem, ClaimService, ClaimDetail, ClaimServiceItem, ClaimServiceService
from medical.models import Item, Service
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from .apps import ClaimConfig


def process_child_relation(user, data_children, claim_id, children, create_hook):
    claimed = 0
    from core.utils import TimeUtils
    if __check_if_maximum_amount_overshoot(data_children, children):
        raise ValidationError(_("mutation.claim_item_service_maximum_amount_overshoot"))
    for data_elt in data_children:
        if ClaimConfig.native_code_for_services == False:
            if create_hook == service_create_hook:
                claimed += calcul_amount_service(data_elt)
            else:
                claimed += data_elt['qty_provided'] * data_elt['price_asked']
        else:
            claimed += data_elt['qty_provided'] * data_elt['price_asked']

        elt_id = data_elt.pop('id') if 'id' in data_elt else None
        if elt_id:
            # elt has been historized along with claim historization
            elt = children.get(id=elt_id)
            [setattr(elt, k, v) for k, v in data_elt.items()]
            elt.validity_from = TimeUtils.now()
            elt.audit_user_id = user.id_for_audit
            elt.claim_id = claim_id
            elt.validity_to = None
            if create_hook == service_create_hook:
                service_update_hook(elt.claim_id, data_elt)

            elt.save()
        else:
            data_elt['validity_from'] = TimeUtils.now()
            data_elt['audit_user_id'] = user.id_for_audit
            # Ensure claim id from func argument will be assigned
            data_elt.pop('claim_id', None)
            # Should entered claim items/services have status passed assigned?
            # Status is mandatory field, and it doesn't have default value in model
            data_elt['status'] = ClaimDetail.STATUS_PASSED
            create_hook(claim_id, data_elt)

    return claimed


def calcul_amount_service(elt):
    totalClaimed = elt['price_asked'] * elt['qty_provided']
    if len(elt['service_item_set']) != 0 and len(elt['service_service_set']) != 0:
        totalClaimed = 0
        for service_item_set in elt['service_item_set']:
            if "qty_asked" in service_item_set:
                if not (math.isnan(service_item_set["qty_asked"])):
                    totalClaimed += service_item_set['qty_asked'] * service_item_set['price_asked']
        for service_service_set in elt['service_service_set']:
            if "qty_asked" in service_service_set:
                if not (math.isnan(service_service_set["qty_asked"])):
                    totalClaimed += service_service_set['qty_asked'] * service_service_set['price_asked']
    return totalClaimed


def __check_if_maximum_amount_overshoot(data_children, children):
    is_overshoot = False
    for entity in data_children:
        quantity = entity.get('qty_provided')
        maximum_amount = None

        if children.model == ClaimItem:
            current_item = Item.objects.get(id=entity['item_id'], validity_to__isnull=True)
            maximum_amount = int(current_item.maximum_amount) if current_item.maximum_amount else None
        elif children.model == ClaimService:
            current_service = Service.objects.get(id=entity['service_id'], validity_to__isnull=True)
            maximum_amount = int(current_service.maximum_amount) if current_service.maximum_amount else None

        if maximum_amount is not None and (quantity > maximum_amount):
            is_overshoot = True
            break

    return is_overshoot


def item_create_hook(claim_id, item):
    # TODO: investigate 'availability' is mandatory,
    # but not in UI > always true?
    item['availability'] = True
    ClaimItem.objects.create(claim_id=claim_id, **item)


def service_create_hook(claim_id, service):
    service_item_set = service.pop('service_item_set', None)
    service_service_set = service.pop('service_service_set', None)
    ClaimServiceId = ClaimService.objects.create(claim_id=claim_id, **service)
    if (service_item_set):
        for serviceL in service_item_set:
            if "qty_asked" in serviceL:
                if (math.isnan(serviceL["qty_asked"])):
                    serviceL["qty_asked"] = 0
            itemId = Item.objects.filter(code=serviceL["sub_item_code"]).first()
            ClaimServiceItem.objects.create(
                item=itemId,
                claim_service=ClaimServiceId,
                qty_displayed=serviceL["qty_asked"],
                qty_provided=serviceL["qty_provided"],
                price_asked=serviceL["price_asked"],
            )

    if (service_service_set):
        for service_service in service_service_set:
            if "qty_asked" in service_service:
                if (math.isnan(service_service["qty_asked"])):
                    service_service["qty_asked"] = 0
            serviceId = Service.objects.filter(code=service_service["sub_service_code"]).first()
            ClaimServiceService.objects.create(
                service=serviceId,
                claim_service=ClaimServiceId,
                qty_displayed=service_service["qty_asked"],
                qty_provided=service_service["qty_provided"],
                price_asked=service_service["price_asked"],
            )


def service_update_hook(claim_id, service):
    service_item_set = service.get("service_item_set")
    service_service_set = service.get("service_service_set")
    service.pop('service_item_set', None)
    service.pop('service_service_set', None)
    ClaimServiceId = ClaimService.objects.filter(claim=claim_id, service=service["service_id"]).first()
    if (service_item_set):
        for serviceL in service_item_set:
            if "qty_asked" in serviceL:
                if (math.isnan(serviceL["qty_asked"])):
                    serviceL["qty_asked"] = 0
            itemId = Item.objects.filter(code=serviceL["sub_item_code"]).first()
            claimServiceItemId = ClaimServiceItem.objects.filter(
                item=itemId,
                claim_service=ClaimServiceId
            ).first()
            claimServiceItemId.qty_displayed = serviceL["qty_asked"]
            claimServiceItemId.save()

    if (service_service_set):
        for service_service in service_service_set:
            if "qty_asked" in service_service:
                if (math.isnan(service_service["qty_asked"])):
                    service_service["qty_asked"] = 0
            serviceId = Service.objects.filter(code=service_service["sub_service_code"]).first()
            claimServiceServiceId = ClaimServiceService.objects.filter(
                service=serviceId,
                claim_service=ClaimServiceId
            ).first()
            claimServiceServiceId.qty_displayed = service_service["qty_asked"]
            claimServiceServiceId.save()


def process_items_relations(user, claim, items):
    return process_child_relation(user, items, claim.id, claim.items, item_create_hook)


def process_services_relations(user, claim, services):
    return process_child_relation(user, services, claim.id, claim.services, service_create_hook)


def autogenerate_nepali_claim_code(config):
    code_length = config.get('code_length')
    if not code_length and type(code_length) is not int:
        raise ValueError("Invalid config for `autogenerate_nepali_claim_code`, expected `code_length` value")
    prefix = __get_current_nepali_fiscal_year_code()
    last_claim = Claim.objects.filter(validity_to__isnull=True, code__icontains=prefix)
    code = 0
    if last_claim:
        code = int(last_claim.latest('code').code[-code_length:])
    return prefix + str(code + 1).zfill(code_length)


def __get_current_nepali_fiscal_year_code():
    import nepali_datetime
    current_date = nepali_datetime.date.today()

    if current_date.month < 4:
        current_year = nepali_datetime.date.today().year - 1
    else:
        current_year = nepali_datetime.date.today().year

    year_code = str(current_year) + "-" + str(current_year + 1)[-3:] + "-"
    return year_code


def get_queryset_valid_at_date(queryset, date):
    filtered_qs = queryset.filter(
        validity_to__gte=date,
        validity_from__lte=date
    )
    if filtered_qs.exists():
        return filtered_qs
    return queryset.filter(validity_from__lte=date, validity_to__isnull=True)



# Generated by Django 3.2.18 on 2023-04-04 14:11

from django.db import migrations
from core.models import RoleRight, Role


def forwards_func(apps, schema_editor):
    # Delete enrollment officer right to interact with claim data
    # Enrollment officer is predefined system role with id 1
    eo_roles = Role.objects.filter(is_system=1, validity_to__isnull=True)
    # Prohibited Roles ID are gql_query_claims_perms (111001)
    # and gql_mutation_deliver_claim_feedback_perms (111009)
    rights_id = [111001, 111009]
    for right_id in rights_id:
        RoleRight.objects.filter(
            role__in=eo_roles,
            right_id=right_id,
            validity_to__isnull=True
        ).delete()


def reverse_func(apps, schema_editor):
    eo_roles = Role.objects.filter(is_system=1, validity_to__isnull=True)
    rights_id = [111001, 111009]
    for eo_role in v:
        for right_id in rights_id:
            RoleRight(
                role_id=eo_role.id,
                right_id=right_id,
                audit_user_id=None,
            ).save()


class Migration(migrations.Migration):

    dependencies = [
        ("claim", "0014_change_code_limit"),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]
